import { useState, useEffect, useCallback } from 'react';

var completionTasks = new Map();
var isCapturing = false;
function taskCapture(task) {
  if (isCapturing) {
    task();
  } else {
    isCapturing = true;
    task();
    completionTasks.forEach(function (task) {
      return task();
    });
    completionTasks.clear();
    isCapturing = false;
  }
}
function taskComplete(key, task) {
  if (isCapturing) {
    if (task) {
      completionTasks.set(key, task);
    } else {
      completionTasks["delete"](key);
    }
  } else {
    task === null || task === void 0 ? void 0 : task();
  }
}
function isTaskCapturing() {
  return isCapturing;
}

function getFunction(selector) {
  return selector.get();
}

function setFunction(state, value) {
  if (!isTaskCapturing()) {
    throw new Error('State cannot be set asynchronously within an action');
  }

  state.set(value);
}

function dispatchFunction(action, value) {
  action.dispatch(value);
}

function createAction(setter) {
  function dispatch(value) {
    taskCapture(function () {
      return setter({
        value: value,
        get: getFunction,
        set: setFunction,
        dispatch: dispatchFunction
      });
    });
  }

  return {
    dispatch: dispatch
  };
}

function bucket(value) {
  return {
    value: value
  };
}

var count = 0;
function generateID() {
  return "id".concat(count++);
}

function createState(initialValue) {
  var stateID = generateID();
  var observers = new Set();
  var current = null;

  function get() {
    var _current;

    current = (_current = current) !== null && _current !== void 0 ? _current : bucket(initialValue instanceof Function ? initialValue() : initialValue);
    return current.value;
  }

  function set(value) {
    var previous = observers.size > 0 ? bucket(get()) : null;
    var next = bucket(value instanceof Function ? value(get()) : value);
    current = next;

    if (previous && next.value !== previous.value) {
      var currentObservers = Array.from(observers);
      taskComplete(stateID, function () {
        return currentObservers.forEach(function (observer) {
          return observer(next.value, previous.value);
        });
      });
    }
  }

  function observe(observer) {
    var wrappedObserver = function wrappedObserver(value, oldValue) {
      return observer(value, oldValue);
    };

    observers.add(wrappedObserver);
    return function () {
      return observers["delete"](wrappedObserver);
    };
  }

  return {
    get: get,
    set: set,
    observe: Object.assign(observe, {
      observers: observers
    })
  };
}

function createMutableState(initialValue) {
  var proxyState = createState(function () {
    return bucket(initialValue instanceof Function ? initialValue() : initialValue);
  });
  var observers = proxyState.observe.observers;

  function get() {
    return proxyState.get().value;
  }

  function set(value) {
    proxyState.set(bucket(value instanceof Function ? value(get()) : value));
  }

  function observe(observer) {
    return proxyState.observe(function (_ref, _ref2) {
      var value = _ref.value;
      var oldValue = _ref2.value;
      observer(value, oldValue);
    });
  }

  return {
    get: get,
    set: set,
    observe: Object.assign(observe, {
      observers: observers
    })
  };
}

var UNINITIALIZED_VALUE = Symbol('Uninitialized value');

function enforceValue(value) {
  if (value === UNINITIALIZED_VALUE) {
    throw new Error("Trying to access selector value before it's computed");
  }

  return value;
}

function createSelector(getter) {
  var proxyState = createState(UNINITIALIZED_VALUE);
  var observers = proxyState.observe.observers;
  var dependencies = new Set();
  var observed = new Set();
  var isStale = true;
  var isObserved = false;

  function getFunction(selector) {
    if (!dependencies.has(selector)) {
      dependencies.add(selector);
      observed.add(selector.observe(observeDependent));
    }

    return selector.get();
  }

  function observeDependent() {
    observed.forEach(function (unobserve) {
      return unobserve();
    });
    observed.clear();
    dependencies.clear();
    isStale = true;

    if (isObserved) {
      computeValue();
    }
  }

  function computeValue() {
    if (isStale) {
      proxyState.set(function () {
        return getter({
          get: getFunction
        });
      });
      isStale = false;
    }
  }

  function get() {
    computeValue();
    return enforceValue(proxyState.get());
  }

  function observe(observer) {
    computeValue();
    isObserved = true;
    var unobserve = proxyState.observe(function (value, oldValue) {
      return observer(enforceValue(value), enforceValue(oldValue));
    });
    return function () {
      unobserve();
      isObserved = observers.size > 0;
    };
  }

  return {
    get: get,
    observe: Object.assign(observe, {
      observers: observers
    })
  };
}

function createProxyState(getter, setter) {
  var proxySelector = createSelector(getter);
  var proxyAction = createAction(setter);

  function set(value) {
    proxyAction.dispatch(value instanceof Function ? value(proxySelector.get()) : value);
  }

  return {
    get: proxySelector.get,
    set: set,
    observe: proxySelector.observe
  };
}

function _typeof(obj) {
  "@babel/helpers - typeof";

  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _createForOfIteratorHelper(o, allowArrayLike) {
  var it;

  if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;

      var F = function () {};

      return {
        s: F,
        n: function () {
          if (i >= o.length) return {
            done: true
          };
          return {
            done: false,
            value: o[i++]
          };
        },
        e: function (e) {
          throw e;
        },
        f: F
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var normalCompletion = true,
      didErr = false,
      err;
  return {
    s: function () {
      it = o[Symbol.iterator]();
    },
    n: function () {
      var step = it.next();
      normalCompletion = step.done;
      return step;
    },
    e: function (e) {
      didErr = true;
      err = e;
    },
    f: function () {
      try {
        if (!normalCompletion && it.return != null) it.return();
      } finally {
        if (didErr) throw err;
      }
    }
  };
}

function filter(object, fn) {
  if (object instanceof Set) {
    return new Set(Array.from(object).filter(function (value) {
      return fn(value);
    }));
  }

  if (object instanceof Map) {
    return new Map(Array.from(object).filter(function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          key = _ref2[0],
          value = _ref2[1];

      return fn(value, key);
    }));
  }

  throw new TypeError('Object must be a set or map instance');
}

function isSelector(obj) {
  return _typeof(obj) === 'object' && obj != null && obj.get instanceof Function && obj.observe instanceof Function;
}

function isValueInRange(value, range) {
  if (range instanceof Function) {
    return range(value);
  }

  if (Array.isArray(range)) {
    return range.includes(value);
  }

  if (range instanceof Set || range instanceof Map) {
    return range.has(value);
  }

  var _iterator = _createForOfIteratorHelper(range),
      _step;

  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var validValue = _step.value;

      if (validValue === value) {
        return true;
      }
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }

  return false;
}

function shallowEqual(objectA, objectB) {
  if (objectA instanceof Map && objectB instanceof Map) {
    return objectA.size === objectB.size && Array.from(objectA).reduce(function (isEqual, _ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          key = _ref2[0],
          value = _ref2[1];

      return isEqual && objectB.has(key) && objectB.get(key) === value;
    }, true);
  }

  if (objectA instanceof Set && objectB instanceof Set) {
    return objectA.size === objectB.size && Array.from(objectA).reduce(function (isEqual, key) {
      return isEqual && objectB.has(key);
    }, true);
  }

  throw new TypeError('Object must be a set or map instance');
}

function createFactory(fn) {
  var keyRange = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function () {
    return true;
  };
  var keysState = createMutableState(new Set());
  var keysSelector = createSelector(function (_ref) {
    var get = _ref.get;
    return new Set(get(keysState));
  });
  var keyRangeSelector = isSelector(keyRange) ? keyRange : createSelector(function () {
    return keyRange;
  });
  var valuesState = createMutableState(new Map());
  var factoryState = createProxyState(function (_ref2) {
    var get = _ref2.get;
    return new Map(get(valuesState));
  }, function (_ref3) {
    var newValues = _ref3.value,
        get = _ref3.get,
        set = _ref3.set;
    var newKeys = new Set(newValues.keys());
    var oldValues = get(valuesState);
    var oldKeys = get(keysState);
    var hasNewKeys = !shallowEqual(newKeys, oldKeys);
    var hasNewValues = hasNewKeys || !shallowEqual(newValues, oldValues);

    if (hasNewValues) {
      set(valuesState, new Map(newValues));
    }

    if (hasNewKeys) {
      set(keysState, newKeys);
    }
  });
  var addAction = createAction(function (_ref4) {
    var addValues = _ref4.value,
        set = _ref4.set;
    set(keysState, function (keys) {
      return addValues.forEach(function (_, key) {
        return keys.add(key);
      }), keys;
    });
    set(valuesState, function (values) {
      return addValues.forEach(function (value, key) {
        return values.set(key, value);
      }), values;
    });
  });
  var deleteAction = createAction(function (_ref5) {
    var deleteKeys = _ref5.value,
        set = _ref5.set;
    set(keysState, function (keys) {
      return deleteKeys.forEach(function (key) {
        return keys["delete"](key);
      }), keys;
    });
    set(valuesState, function (values) {
      return deleteKeys.forEach(function (key) {
        return values["delete"](key);
      }), values;
    });
  });
  keyRangeSelector.observe(function (range) {
    var allKeys = keysState.get();
    var deleteKeys = filter(allKeys, function (key) {
      return !isValueInRange(key, range);
    });
    deleteAction.dispatch(deleteKeys);
  });

  function factory(key) {
    var _allValues$get;

    if (!isValueInRange(key, keyRangeSelector.get())) {
      throw new RangeError("Factory key \"".concat(key, "\" is out of bounds"));
    }

    var allValues = valuesState.get();
    var value = (_allValues$get = allValues.get(key)) !== null && _allValues$get !== void 0 ? _allValues$get : fn(key);

    if (!allValues.has(key)) {
      var addValues = new Map([[key, value]]);
      addAction.dispatch(addValues);
    }

    return value;
  }

  return Object.assign(factory, factoryState, {
    keys: keysSelector
  });
}

function map(object, fn) {
  if (object instanceof Set) {
    return new Set(Array.from(object, function (value) {
      return fn(value);
    }));
  }

  if (object instanceof Map) {
    return new Map(Array.from(object, function (_ref) {
      var _ref2 = _slicedToArray(_ref, 2),
          key = _ref2[0],
          value = _ref2[1];

      return [key, fn(value, key)];
    }));
  }

  throw new TypeError('Object must be a set or map instance');
}

function createSelectorFactory(fn, keyRange) {
  var proxyFactory = createFactory(function (key) {
    var getterOrSelector = fn(key);
    return isSelector(getterOrSelector) ? getterOrSelector : createSelector(getterOrSelector);
  }, keyRange);
  var factorySelector = createSelector(function (_ref) {
    var get = _ref.get;
    return map(get(proxyFactory), function (state) {
      return get(state);
    });
  });

  function factory(key) {
    return proxyFactory(key);
  }

  return Object.assign(factory, proxyFactory, factorySelector);
}

function isState(obj) {
  return isSelector(obj) && typeof obj.set === 'function';
}

function createStateFactory(fn, keyRange) {
  var proxyFactory = createFactory(function (key) {
    var initialValueOrState = fn(key);
    return isState(initialValueOrState) ? initialValueOrState : createState(initialValueOrState);
  }, keyRange);
  var factoryState = createProxyState(function (_ref) {
    var get = _ref.get;
    return map(get(proxyFactory), function (state) {
      return get(state);
    });
  }, function (_ref2) {
    var values = _ref2.value,
        set = _ref2.set;
    values.forEach(function (value, key) {
      return set(proxyFactory(key), value);
    });
    set(proxyFactory, map(values, function (_, key) {
      return proxyFactory(key);
    }));
  });

  function factory(key) {
    return proxyFactory(key);
  }

  return Object.assign(factory, proxyFactory, factoryState);
}

function asError(error) {
  return error instanceof Error ? error : new Error(error != null ? String(error) : undefined);
}

function useAsync(fn) {
  var deps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

  var _useState = useState({
    complete: false
  }),
      _useState2 = _slicedToArray(_useState, 2),
      state = _useState2[0],
      setState = _useState2[1];

  useEffect(function () {
    var cancelled = false;
    setState({
      complete: false
    });
    fn().then(function (value) {
      return !cancelled && setState({
        value: value,
        complete: true
      });
    }, function (reason) {
      return !cancelled && setState({
        error: asError(reason),
        complete: true
      });
    });
    return function () {
      cancelled = true;
    };
  }, deps);
  return state;
}

function useSelector(sharedSelector) {
  var value = sharedSelector.get();

  var _useState = useState(0),
      _useState2 = _slicedToArray(_useState, 2),
      setObserveCounter = _useState2[1];

  useEffect(function () {
    return sharedSelector.observe(function () {
      return setObserveCounter(function (count) {
        return count + 1;
      });
    });
  }, [sharedSelector]);
  return value;
}

function useAsyncSelector(selector) {
  var promise = useSelector(selector);

  var _useAsync = useAsync(function () {
    return promise;
  }, [promise]),
      value = _useAsync.value,
      error = _useAsync.error;

  return [value, error];
}

function useSharedState(state) {
  var value = useSelector(state);
  return [value, state.set];
}

function useAsyncState(state) {
  var _useSharedState = useSharedState(state),
      _useSharedState2 = _slicedToArray(_useSharedState, 2),
      promise = _useSharedState2[0],
      setPromise = _useSharedState2[1];

  var _useAsync = useAsync(function () {
    return promise;
  }, [promise]),
      value = _useAsync.value,
      error = _useAsync.error;

  var setValue = useCallback(function (value) {
    return setPromise(value instanceof Function ? function (promise) {
      return promise.then(value);
    } : Promise.resolve(value));
  }, [setPromise]);
  return [value, setValue, error];
}

export { createAction, createFactory, createMutableState, createProxyState, createSelector, createSelectorFactory, createState, createStateFactory, isSelector, isState, isTaskCapturing, taskCapture, taskComplete, useAsyncSelector, useAsyncState, useSelector, useSharedState };
